package middleware.agents;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.LinkedBlockingQueue;
import middleware.message.MetaAgentMessage;
import middleware.message.SystemErrorMessage;
import middleware.message.wrapper.MetaAgentWrapper;
import middleware.observer.Observer;
import middleware.observer.Subject;

/**
 * MetaAgent
 *
 * @author JLScott1999
 */
public abstract class MetaAgent extends LinkedBlockingQueue implements Runnable, Subject
{

    private AgentAddress address;
    private Thread thread;
    private boolean running = true;
    private ArrayList<MetaAgent> metaAgent = new ArrayList();
    private List<Observer> listOfObservers = new ArrayList<>();

    /**
     * Creates a <i>MetaAgent</i> with specified properties
     * @param queueSize the size of the blocking queue
     * @param addressString address of the MetaAgent
     */
    public MetaAgent(int queueSize, String addressString)
    {
        super(queueSize);
        this.address = new AgentAddress(addressString);
    }

    /**
     * Runs the MessageHandler once thread is started
     */
    @Override
    public void run()
    {
        while (running)
        {
            try
            {
                msgHandler(super.take());
            }
            catch (InterruptedException ex)
            {
                System.out.println("Exiting");
            }
        }
    }

    /**
     * Starts thread
     */
    public void start()
    {
        if (thread == null)
        {
            running = true;
            thread = new Thread(this);
            thread.start();
        }
    }

    /**
     * Stops thread
     */
    public void stop()
    {
        if (thread != null)
        {
            running = false;
            thread.interrupt();
            thread = null;
        }
    }

    /**
     * Gets <i>MetaAgent</i>'s address
     * @return MetaAgent's Address
     */
    public AgentAddress getAddress()
    {
        return address;
    }

    /**
     * Sets <i>MetaAgent</i>'s address
     * @param addressString MetaAgent's Address
     */
    public void setAddress(String addressString)
    {
        this.address = new AgentAddress(addressString);
    }

    /**
     * Attaches an observer to the MetaAgent
     * @param observer observer to be attached (e.g. NodeListener)
     */
    @Override
    public void attach(Observer observer) {
        listOfObservers.add(observer);
    }

    /**
     * Detaches an observer from the MetaAgent (if it is attached)
     * @param observer observer to be detached
     */
    @Override
    public void detach(Observer observer) {
        if(listOfObservers.contains(observer)){
            listOfObservers.remove(observer);
        }
    }

    /**
     * Updates attached observers with a message
     * @param msg the message to be sent to the observers
     * @return whether observers were successfully updated
     */
    @Override
    public boolean updateObservers(Object msg) {
        boolean result = false;
        for (Observer o : listOfObservers){
            result = o.update(msg);
        }
        return result;
    }

    /**
     * Allows for communication between MetaAgents
     * Adds a MetaAgent
     * @param agent  
     */
    public void addAgent(MetaAgent agent)
    {
        // TODO Generate random address
        if (!metaAgent.contains(agent))
        {
            metaAgent.add(agent);
            agent.addAgent(this);
        }
    }
    
    /**
     * Handles message objects passed to it
     * @param data message
     */
    public void msgHandler(Object data)
    {
        if (data instanceof MetaAgentMessage)
        {
            if (getAddress().getSection(1) != ((MetaAgentMessage) data).getMessage().getRecipient().getSection(1))
            {
                boolean sent = false;
                for (int i = 0; i < metaAgent.size() && !sent; i++)
                {
                    if (metaAgent.get(i).getAddress().toString().equals(this.getAddress().getSection(1) + ".0.0"))
                    {
                        metaAgent.get(i).add(data);
                        sent = true;
                    }
                }
                if(!sent)
                {
                    if (!((MetaAgentMessage) data).getMessage().getSender().getAddress().equals("0.0.0"))
                    {
                        add(MetaAgentWrapper.wrap(new SystemErrorMessage(((MetaAgentMessage) data).getMessage().getRecipient() + " Does not exist?", ((MetaAgentMessage) data).getMessage().getSender().getAddress()), 15));
                    }
                }
            }
            else
            {
                boolean sent = false;
                for (int i = 0; i < metaAgent.size() && !sent; i++)
                {
                    if (metaAgent.get(i).getAddress().getSection(2) == ((MetaAgentMessage) data).getMessage().getRecipient().getSection(2) && metaAgent.get(i).getAddress().getSection(1) == ((MetaAgentMessage) data).getMessage().getRecipient().getSection(1))
                    {
                        metaAgent.get(i).add(data);
                        sent = true;
                    }
                }
                if(!sent)
                {
                for (int i = 0; i < metaAgent.size() && !sent; i++)
                {
                    if (metaAgent.get(i).getAddress().getSection(1) == this.getAddress().getSection(1))
                    {
                        metaAgent.get(i).add(data);
                        sent = true;
                    }
                }
                }
                if(!sent)
                {
                    if (!((MetaAgentMessage) data).getMessage().getSender().getAddress().equals("0.0.0"))
                    {
                        add(MetaAgentWrapper.wrap(new SystemErrorMessage(((MetaAgentMessage) data).getMessage().getRecipient() + " Does not exist?", ((MetaAgentMessage) data).getMessage().getSender().getAddress()), 15));
                    }
                }
            }
        }
    }

}
